# Copyright 2021 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Loads and runs a SavedModel generated by :keras_test on the Adult dataset.

This binary tests the model in a clean tensorflow instance (without pre-loaded
functions and resources) and does not require non-canonical tf ops.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import enum
from typing import Dict

from absl import app
from absl import flags
from absl import logging
import numpy as np
import pandas as pd
import tensorflow as tf

# Inject the inference ops only.
from tensorflow_decision_forests.tensorflow.ops.inference import op  # pylint: disable=unused-import

flags.DEFINE_string("model_path", "", "Path to the SavedModel model to load.")

flags.DEFINE_string("dataset_path", "", "Path to adult dataset.")

FLAGS = flags.FLAGS


def main(argv):
  del argv

  logging.info("Load model")
  model = tf.keras.models.load_model(FLAGS.model_path)

  logging.info("Load dataset")
  pd_dataset = load_dataset()
  tf_dataset = dataframe_to_tf_dataset(pd_dataset, "income")

  logging.info("Evaluate model")
  evaluation = model.evaluate(tf_dataset)
  logging.info("evaluation: %s", evaluation)

  logging.info("Done (unload model)")


# The following code is similar to the simple_ml_keras_test.


class Semantic(enum.Enum):
  NUMERICAL = 1
  CATEGORICAL = 2


def dataframe_to_tf_dataset(dataframe: pd.DataFrame,
                            label: str) -> tf.data.Dataset:

  dataset = tf.data.Dataset.from_tensor_slices(
      (dict(dataframe.drop(label, axis=1)), dataframe[label].values))
  return dataset.batch(64)


def load_dataset() -> pd.DataFrame:
  """Loads the dataset in a pandas dataframe."""

  dataset = pd.read_csv(flags.FLAGS.dataset_path)
  label = "income"

  def clean(ds):
    ds[label] = np.where(ds[label] == ">50K", 1, 0)
    return ds

  dataset = clean(dataset)
  return auto_dataset(dataset, label)


def infer_semantic_from_dataframe(dataset: pd.DataFrame) -> Dict[str, Semantic]:
  """Infers the semantic of the columns in a pandas dataframe."""

  semantics = {}
  for col in dataset.columns:
    dtype = dataset[col].dtype
    if dtype in [float, np.int16, np.int32, np.int64, int]:
      semantics[col] = Semantic.NUMERICAL
    elif dtype in [str, object]:
      semantics[col] = Semantic.CATEGORICAL
    else:
      raise Exception(
          "Cannot infer semantic for column \"{}\" with dtype={}".format(
              col, dtype))

  return semantics


def auto_dataset(dataset, label) -> pd.DataFrame:
  """Prepares a dataset object."""

  semantics = infer_semantic_from_dataframe(dataset)

  def extract_by_semantic(semantic):
    return [k for k, s in semantics.items() if s == semantic and k != label]

  categorical_features = extract_by_semantic(Semantic.CATEGORICAL)
  numerical_features = extract_by_semantic(Semantic.NUMERICAL)

  def clean(dataset):
    for key in categorical_features:
      dataset[key] = dataset[key].fillna("")
    return dataset[categorical_features + numerical_features + [label]]

  return clean(dataset)


if __name__ == "__main__":
  app.run(main)
